//---------------------------------------------------------------------------
// 2B_Ping_pong_various_sends.cpp 
//---------------------------------------------------------------------------
#if (defined __linux__) || (defined _AIX) || (defined __APPLE__)
	#include <sys/types.h>
	#include <sys/stat.h>
	#include <unistd.h>
#elif (defined _WIN32) || (defined _WIN64)
	#include <conio.h>
	#include <direct.h>
#endif

#include "mpi.h"
#include <iostream>
using namespace std;

int main(int argc, char* argv[])
{
	int i, j, rank, size, mode, first_process;
	double time;
	int array[10];

	MPI_Request request;
	MPI_Status status;

	MPI_Init(&argc, &argv);

	MPI_Comm_size(MPI_COMM_WORLD, &size);
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);

	if (rank == 0)
		cout << "There are " << size << " processes";

	if (rank == 0) {
		cout << "\n---------------------------------------------------------------";
		cout << "\nSelect a mode:\n  1. MPI_Send\n  2. MPI_Ssend (deadlock)\n"
			<< "  3. MPI_Ssend\n  4. MPI_Issend\n\n  mode = ";
		cin >> mode;
		MPI_Ssend(&mode, 1, MPI_INT, 1, 0, MPI_COMM_WORLD);
	}
	else
		MPI_Recv(&mode, 1, MPI_INT, 0, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);

	if (rank == 0) {
		cout << "\nSelect which process will operate as the first:\n\n"  
			<< "  first process = ";
		cin >> first_process;
		cout << "---------------------------------------------------------------\n";
		MPI_Ssend(&first_process, 1, MPI_INT, 1, 0, MPI_COMM_WORLD);
	}
	else
		MPI_Recv(&first_process, 1, MPI_INT, 0, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);

	for (i = 0; i < 10; i++)
		array[i] = 0;

	if (rank == 0) {
		cout << "\nprocess " << rank << " - Initial values of array are:" << endl;
		for (i = 0; i < 10; i++)
			cout << array[i] << " ";
		cout << endl;
		std::fflush(stdout);
	}

	MPI_Barrier(MPI_COMM_WORLD);
	time = -MPI_Wtime();

	if (rank != first_process)
		for (i = 0; i < 1000000; i++)       // zdrzujeme second_proces
			for (j = 0; j < 1000000; j++)       
				if (i <= j)
					i = j;

	cout << "\nprocess " << rank << " is prepared" << endl;
	std::fflush(stdout);

	switch (mode) {
	case 1:
		for (i = 0; i < 1000000; i++) {
			if (rank == 0) {
				MPI_Send(array, 10, MPI_INT, 1, 0, MPI_COMM_WORLD);

				MPI_Recv(array, 10, MPI_INT, 1, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
			}
			if (rank == 1) {
				for (j = 0; j < 10; j++)
					array[j]++;
				MPI_Send(array, 10, MPI_INT, 0, 0, MPI_COMM_WORLD);

				MPI_Recv(array, 10, MPI_INT, 0, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
			}
		} // array[j] = 500 000, prikazy v if-och sa vykonaju hned a spolu naraz,
		// necaka sa na synchronizaciu s druhym if-om.
		// Ak MPI_Recv(array, 10, MPI_INT, 1, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE) 
		// nenajde na sieti odpovedajuci MPI_Send (= matching), netrapi sa tym a 
		// ide dalej. O synchronizaciu sa musite postarat sami.
		break;
	case 2:
		for (i = 0; i < 1000000; i++) {
			if (rank == 0) {
				cout << "process " << rank << ": i=" << i << ", stage 0" << endl;
				MPI_Ssend(array, 10, MPI_INT, 1, 0, MPI_COMM_WORLD);
				cout << "process " << rank << ": i=" << i << ", stage 1" << endl;
				MPI_Recv(array, 10, MPI_INT, 1, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
				cout << "process " << rank << ": i=" << i << ", stage 2" << endl;
			}
			if (rank == 1) {
				for (j = 0; j < 10; j++)
					array[j]++;
				cout << "process " << rank << ": i=" << i << ", stage 3" << endl;
				MPI_Ssend(array, 10, MPI_INT, 0, 0, MPI_COMM_WORLD);
				cout << "process " << rank << ": i=" << i << ", stage 4" << endl;
				MPI_Recv(array, 10, MPI_INT, 0, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
				cout << "process " << rank << ": i=" << i << ", stage 5" << endl;
			}
		} // DEADLOCK !!!
		break;
	case 3:
		for (i = 0; i < 1000000; i++) {
			if (rank == 0) {
				MPI_Ssend(array, 10, MPI_INT, 1, 0, MPI_COMM_WORLD);

				MPI_Recv(array, 10, MPI_INT, 1, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
			}
			if (rank == 1) {	// proces 1 POCKA na zasielku od procesu 0
				MPI_Recv(array, 10, MPI_INT, 0, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);

				for (j = 0; j < 10; j++)
					array[j]++;
				MPI_Ssend(array, 10, MPI_INT, 0, 0, MPI_COMM_WORLD);
			}
		}
		break;
	case 4:
		for (i = 0; i < 1000000; i++) {
			if (rank == 0) {
				MPI_Issend(array, 10, MPI_INT, 1, 0, MPI_COMM_WORLD, &request);
				// Do_something_else_while Issend_happens();
				// Now wait for send to complete
				MPI_Wait(&request, &status);

				MPI_Irecv(array, 10, MPI_INT, 1, 1, MPI_COMM_WORLD, &request);
				// Do_something_else_while MPI_Irecv_happens();
				// Now wait for receive to complete
				MPI_Wait(&request, &status);
			}
			if (rank == 1) {	// proces 1 POCKA na zasielku od procesu 0
				MPI_Irecv(array, 10, MPI_INT, 0, 0, MPI_COMM_WORLD, &request);
				// Do_something_else_while MPI_Irecv_happens();
				// Now wait for receive to complete
				MPI_Wait(&request, &status);

				for (j = 0; j < 10; j++)
					array[j]++;
				MPI_Issend(array, 10, MPI_INT, 0, 1, MPI_COMM_WORLD, &request);
				// Do_something_else_while Issend_happens();
				// Now wait for send to complete
				MPI_Wait(&request, &status);
			}
		}
		break;

	default: break;
	}

	if (rank == 0) {
		cout << "\nprocess " << rank << " - Final values of array are:" << endl;
		for (i = 0; i < 10; i++)
			cout << array[i] << " ";
	}

	time += MPI_Wtime();
	if (rank == 0)
		printf("\n\nTime = %8.6f s\n", time);

	std::fflush(stdout);
	MPI_Finalize();

	return 0;
}